; -*- Mode: Common-Lisp -*-
(in-package 'flavors)
(provide 'flavors)
(pushnew :APOLLO-FLAVORS *MODULES*)
(eval-when (load eval compile)
  (setf *FEATURES* (delete  :STREAM-FLAVORS *FEATURES*))
  (setf *FEATURES* (delete  :HASH-FLAVORS *FEATURES*))
  (setf *FEATURES* (delete  :VECTOR-FLAVORS *FEATURES*))
  (setf *FEATURES* (delete  :PROGV-FLAVORS *FEATURES*))
  ;;(pushnew :STREAM-FLAVORS *FEATURES*)
  (pushnew :HASH-FLAVORS    *FEATURES*)
  (pushnew :VECTOR-FLAVORS  *FEATURES*)
  ;;(pushnew :PROGV-FLAVORS *FEATURES*)
  )

(export '(mtrace muntrace))

(proclaim '(inline send iv-name ordinary-print-instance keyword))

(defvar *UNBOUND* (intern "<UNBOUND>"))
(defvar *ALL-FLAVOR-NAMES* nil)
(defvar *ALL-SLOT-REFERENCES* nil)
(defvar *COMPILE-COMBINED-WHEN*      '(:RECOMPILE :COMPILE :SEND :GET-HANDLER-FOR))
(defvar *COMPILE-UNINSTANCIATED-WHEN* '(:RECOMPILE :COMPILE))
(defvar *FAILURE-RESULT* :ERROR)
(defvar *FLAVORS* (make-hash-table :test #'eq))
(defvar *INSIDE* nil)
(defvar *INSTANCE-INDICES* nil)
(defvar *INSTANCE-VARIABLES* nil)
(defvar *NUMBERS* nil)
(defvar *PRINT-NAMES* (make-hash-table :test #'eq))
(defvar *RESET-COMBINED-WHEN*         '(:RECOMPILE :SETDAEMON :DEFFLAVOR))
(defvar *SEND-INIT-MESSAGE-P* t)
(defvar *TIME* nil)
(defvar *WHOPPER-ARGS* nil)
(defvar *WHOPPER-CONTINUE-HOOK* nil)
(defvar SELF 'SELF)

(defadvice (lucid::output-object hash-object) (object &rest args)
  (apply-advice-continue 
    (if (symbolp object) 
	(gethash object *PRINT-NAMES* object) 
	object)
    args))

(defadvice (LUCID::CONSTRUCT-HAIRY-REDEF-MSG hash-object) (object &rest args)
  (let ((o (gethash object *PRINT-NAMES*)))
    ;;; Oy, Oy, Oy...
    (if o (setq object (prin1-to-string o)))
    (apply-advice-continue object args)))

(defun iv-name (s) 
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (if (consp s) (car s) s))

(defun keyword (x)
  (intern (symbol-name x) (find-package "KEYWORD")))

(defstruct (method-token (:print-function print-method-token))
  name
  daemon
  message)

(defun print-method-token (method-token &optional (stream *STANDARD-OUTPUT*) &rest ignore)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (ignore ignore))
  (format stream "(:METHOD ~S ~S ~S)" 
	  (method-token-name   method-token) 
	  (method-token-daemon method-token)
	  (method-token-message method-token)))

(defmacro mtrace (&rest args)
  (labels ((fix-arg (trace-arg)
	     (if (and (consp trace-arg) (not (keywordp (car trace-arg))))
		 (if (and (consp (car trace-arg)) (not (keywordp (caar trace-arg))))
		     (rplaca trace-arg (mapcar #'method-symbol (car trace-arg)))
		     (rplaca trace-arg (method-symbol (car trace-arg))))
		 (method-symbol trace-arg))))
    (let ((*FAILURE-RESULT* :TOKEN))
      `(trace ,@(mapcar #'fix-arg args)))))

(defmacro muntrace (&rest args)
  (labels ((fix-arg (trace-arg)
	     (if (and (consp trace-arg) (not (keywordp (car trace-arg))))
		 (if (and (consp (car trace-arg)) (not (keywordp (caar trace-arg))))
		     (rplaca trace-arg (mapcar #'method-symbol (car trace-arg)))
		     (rplaca trace-arg (method-symbol (car trace-arg))))
		 (method-symbol trace-arg))))
    (let ((*FAILURE-RESULT* :TOKEN))
      `(untrace ,@(mapcar #'fix-arg args)))))

(defmacro with-send-environment ((&optional instance) . body)
  #-PROGV-FLAVORS
  `(let ((*INSTANCE-VARIABLES* (instance-variables ,instance))
	 (*INSTANCE-INDICES*   (instance-indices ,instance)))
     (declare (special *INSTANCE-VARIABLES* *INSTANCE-INDICES*))
     ,@body)
  #+PROGV-FLAVORS
  `(progv (unless (eq (instance-flavor ,instance) *INSIDE*) (instance-indices ,instance)) *NUMBERS*
     (let ((*INSIDE* (instance-flavor ,instance)))
       ,@body)))

(eval-when (load eval compile)  
(defun with-flavor-ivs (ivs-used body &optional dynamic-p)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  ivs-used
  #+(and VECTOR-FLAVORS (not PROGV-FLAVORS))
  (flet ((index-static  (iv) `(gethash ',iv *instance-indices*))
	 (index-dynamic (iv) `(gethash ,iv *instance-indices*)))
    (when ivs-used
      (setq body 
	    `(((lambda ,ivs-used 
		 (declare (fixnum ,@ivs-used) (inline svref lucid::set-svref))
		 ,@body)
	       ,@(mapcar (if dynamic-p #'index-dynamic #'index-static) ivs-used))))))
  body)
)

(defmacro with-flavor-ivs-dynamic ((instance &rest ivs-used) &rest body)
  `(with-send-environment (,instance)
     ,@(with-flavor-ivs ivs-used body 'dynamic)))

(defmacro ivs (instance) (declare (ignore instance)) '*INSTANCE-VARIABLES*)

(defmacro variable-to-reference (var)
  #+HASH-FLAVORS #-VECTOR-FLAVORS `(quote ,var)
  #+HASH-FLAVORS #+VECTOR-FLAVORS var
  #+PROGV-FLAVORS
  (let ((r (intern (format nil "~A:~A" (package-name (symbol-package var)) var) "FLAVORS")))
    (unless (member r *ALL-SLOT-REFERENCES* :test #'eq)
      (proclaim `(special ,r))
      (push r *ALL-SLOT-REFERENCES*))
    r))

(defmacro set-variable-list (instance size)
  size
  #+VECTOR-FLAVORS                
  `(setf (instance-variables ,instance) (make-array ,size :initial-element *UNBOUND*))
  #-VECTOR-FLAVORS #-HASH-FLAVORS 
  `(setf (instance-variables ,instance) ,instance)
  #-VECTOR-FLAVORS #+HASH-FLAVORS 
  `(setf (instance-variables ,instance) (make-hash-table :test #'eq :size ,size)))

(defmacro getvariable-from-instance (instance variable &optional (default *UNBOUND*))
  #+VECTOR-FLAVORS (declare (ignore default))
  #+VECTOR-FLAVORS 
  `(svref (ivs ,instance) (gethash ',(variable-to-reference variable) *INSTANCE-INDICES*))
  #-VECTOR-FLAVORS #+HASH-FLAVORS 
  `(gethash ,(variable-to-reference variable) (ivs ,instance) ,default)
  #-HASH-FLAVORS #-VECTOR-FLAVORS variable)

(defmacro setvariable-from-instance ((instance variable) value) 
  #+VECTOR-FLAVORS 
  `(setf (svref (ivs ,instance) (gethash ',(variable-to-reference variable) *INSTANCE-INDICES*)) ,value)
  #-VECTOR-FLAVORS #+HASH-FLAVORS `(setf (gethash ,(variable-to-reference variable) (ivs ,instance)) ,value)
  #-VECTOR-FLAVORS #-HASH-FLAVORS `(setf ,(variable-to-reference variable) ,value))

(defsetf getvariable-from-instance (instance variable) (value) 
  `(setvariable-from-instance (,instance ,variable) ,value))

(defmacro getvariable (instance variable &optional (default *UNBOUND*))
  #+VECTOR-FLAVORS (declare (ignore default))
  #+VECTOR-FLAVORS `(svref (ivs ,instance) ,(variable-to-reference variable))
  #-VECTOR-FLAVORS #+HASH-FLAVORS `(gethash ,(variable-to-reference variable) (ivs ,instance) ,default)
  #-HASH-FLAVORS #-VECTOR-FLAVORS variable)

(defmacro setvariable ((instance variable) value) 
  #+VECTOR-FLAVORS `(setf (svref (ivs ,instance) ,(variable-to-reference variable)) ,value)
  #-VECTOR-FLAVORS #+HASH-FLAVORS `(setf (gethash ,(variable-to-reference variable) (ivs ,instance)) ,value)
  #-VECTOR-FLAVORS #-HASH-FLAVORS `(setf ,(variable-to-reference variable) ,value))

(defsetf getvariable (instance variable) (value) `(setvariable (,instance ,variable) ,value))

(defmacro getvariable-dynamic (instance variable)
  #+VECTOR-FLAVORS `(svref (ivs ,instance) (variable-to-reference ,variable))
  #-VECTOR-FLAVORS #+HASH-FLAVORS `(getvariable ,instance ,variable)
  #-HASH-FLAVORS #-VECTOR-FLAVORS `(getvariable ,instance ,variable))

(defmacro setvariable-dynamic ((instance variable) value)
  #+VECTOR-FLAVORS `(setf (svref (ivs ,instance) (variable-to-reference ,variable)) ,value)
  #-VECTOR-FLAVORS #+HASH-FLAVORS `(setvariable (,instance ,variable) ,value)
  #-VECTOR-FLAVORS #-HASH-FLAVORS `(setvariable (,instance ,variable) ,value))

(defmacro mapvariables (instance function)
  (let ((key (gensym)))
    `(dolist (,key (flavor-variable-names (instance-flavor ,instance)))
       (funcall ,function ,key (getvariable ,instance ,key)))))

(defmacro lexpr-args (args)
  `(let ((a (copy-list ,args)))
     (let ((last (last a)))
       (rplacd last (cdar last))
       (rplaca last (caar last)))
     a))

#+STREAM-FLAVORS
(lucid::define-stream-type instance
    (lucid::basic-stream)
  (flavor
    number 
    indices
    variables)
  (:print-function print-instance)
  (:predicate instancep)
  (:constructor %create-instance))

#-STREAM-FLAVORS
(defstruct (instance
	     (:print-function print-instance)
	     (:predicate instancep)
	     (:constructor %create-instance))
  flavor
  number 
  indices
  variables)
  
(defstruct (method (:print-function print-method))
  flavor
  name
  whopper
  wrapper
  before
  primary
  after
  combined)

(defstruct (flavor (:print-function print-flavor))
  all-initable-instance-variables
  all-init-keywords
  all-required-flavors
  all-required-init-keywords
  all-required-instance-variables
  all-required-methods
  all-variables
  apply-inside-self
  combination-list
  components
  default-init-plist
  dependants
  gettable-instance-variables
  included-flavors
  init-keywords
  indices
  initable-instance-variables
  (instance-number 0)
  instance-variables
  local-variables
  method-combination-modifier
  method-combination-type
  methods
  name
  no-vanilla-flavor
  operations
  required-flavors
  required-init-keywords
  required-instance-variables
  required-methods
  settable-instance-variables
  )

(defadvice (type-of instance-type-of) (x)
  (let ((type (advice-continue x)))
    (if (eq type 'instance)
	(flavor-name (instance-flavor x))
	type)))

(defmacro flavor-instanciated (flavor) `(> (flavor-instance-number ,flavor) 0))

(defun ordinary-print-instance (instance &optional (stream *STANDARD-OUTPUT*) prindepth slashify-p)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (ignore prindepth slashify-p))
  (format stream "#<Instance ~A ~S>" 
	  (flavor-name (instance-flavor instance))
	  (instance-number instance)))

(defun print-instance (instance stream print-level)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (ignore print-level))
  (if (operation-handled-p instance :print-self)
      (send instance :print-self stream print-level)
      (ordinary-print-instance instance stream print-level)))

(defun print-method (method stream print-level)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (ignore print-level))
  (format stream "#<Method-Structure ~A :~A [~A]>"
	  (flavor-name (method-flavor method))
	  (method-name method)
	  (concatenate 'string
		       (if (method-combined method) "C" "")
		       (if (method-wrapper method)  "R" "")
		       (if (method-whopper method)  "W" "")
		       (if (method-before  method)  "B" "")
		       (if (method-after   method)  "A" "")
		       (if (method-primary method)  "P" ""))))

(defun print-flavor (flavor stream print-level)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (ignore print-level))
  (format stream "#<Flavor ~S>" (flavor-name flavor)))

(defmacro flavor-descriptor (type) `(get ,type :FLAVOR))
(defsetf  flavor-descriptor (type) (value) `(setf (get ,type :FLAVOR) ,value))
(defun    eflavor-descriptor (type)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (or  (flavor-descriptor type)
       (progn
	 (cerror "Continue after defining flavor" "No such flavor: ~S" type)
	 (eflavor-descriptor type))))

(defmacro adjust-max-instance-size (size)
  #+PROGV-FLAVORS 
  `(when (< (length *NUMBERS*) ,size)
     (setq *NUMBERS* (make-list size)) ;; This makes sure that the cells are contiguous
     (do ((nl *NUMBERS* (cdr nl))
	  (n 0 (1+ n)))
	 ((endp n) nil)
       (setf (car nl) n)))
  #-PROGV-FLAVORS (declare (ignore size)))

(defmacro combination-slot (flavor all-slot slot)
  `(or (,all-slot ,flavor)
       (setf (,all-slot ,flavor)
	     (remove-duplicates 
	       (mapcan #'(lambda (f) (copy-list (,slot f)))
		       (cons ,flavor (flavor-combination-list ,flavor)))
	       :from-end t))))

(defun initialize-variables (instance flavor options)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let* ((all-initable-instance-variables 
	  (combination-slot 
	    flavor
	    flavor-all-initable-instance-variables
	    flavor-initable-instance-variables))
	 (length (length (the list (flavor-all-variables flavor))))
	 (variables (set-variable-list instance (max 1 length))))
    (setf (instance-indices   instance) (flavor-indices flavor))
    (adjust-max-instance-size length)
    (dolist (s (flavor-all-variables flavor))
      (let ((variable (iv-name s))
	    (value    *UNBOUND*))
	(when (member variable all-initable-instance-variables :test #'eq)
	  (let ((key (keyword variable)))
	    (setq value (getf options key *UNBOUND*))
	    (when (eq value *UNBOUND*)
	      (setq value (getf (flavor-default-init-plist flavor) key *UNBOUND*)))))
	(when (and (consp s) (eq value *UNBOUND*)) (setq value (eval (cadr s))))
	(with-flavor-ivs-dynamic (instance variable)
	  (setvariable-dynamic (instance variable) value))))
    variables))

(defmacro getdaemon (method daemon)
  (case daemon
    (:combined `(method-combined ,method))
    (:wrapper  `(method-wrapper  ,method))
    (:whopper  `(method-whopper  ,method))
    (:before   `(method-before   ,method)) 
    (:after    `(method-after    ,method))
    ((:primary :progn :and :or :list :append :nconc :case)
     `(method-primary  ,method))
    (t (let ((m (gensym)))
	 `(let ((,m ,method))
	    (ecase ,daemon
	      (:combined (method-combined ,m))
	      (:wrapper  (method-wrapper  ,m))
	      (:whopper  (method-whopper  ,m))
	      (:before   (method-before   ,m)) 
	      (:after    (method-after    ,m))
	      ((:primary :progn :and :or :list :append :nconc :case)
	       (method-primary  ,m))))))))

(defmacro setdaemon ((method daemon) value)
  `(let ((*TIME* :SETDAEMON))
     (unless (or (eq ,daemon :COMBINED) (getdaemon ,method ,daemon))
       (fix-combined-method (method-flavor ,method) (method-name ,method) :DO-DEPENDANTS))
     ,(case daemon
	(:combined `(setf (method-combined ,method) ,value))
	(:wrapper  `(setf (method-wrapper  ,method) ,value))
	(:whopper  `(setf (method-whopper  ,method) ,value))
	(:before   `(setf (method-before   ,method) ,value))
	(:after    `(setf (method-after    ,method) ,value))
	((:primary :progn :and :or :list :append :nconc :case)
	 `(setf (method-primary  ,method) ,value))
	(t (let ((m (gensym)))
	     `(let ((,m ,method))
		(ecase ,daemon
		  (:combined (setf (method-combined ,m) ,value))
		  (:wrapper  (setf (method-wrapper  ,m) ,value))
		  (:whopper  (setf (method-whopper  ,m) ,value))
		  (:before   (setf (method-before   ,m) ,value))
		  (:after    (setf (method-after    ,m) ,value))
		  ((:primary :progn :and :or :list :append :nconc :case) 
		   (setf (method-primary  ,m) ,value)))))))))

(defmacro getmethod (flavor method &optional daemon)
  (let ((meth `(gethash ,method (flavor-methods ,flavor))))
    (if daemon
	(let ((m (gensym)))
	  `(let ((,m ,meth)) (when ,m (getdaemon ,m ,daemon))))
	meth)))

(defmacro setmethod ((flavor method &optional daemon) value)
  `(setmethod-function ,flavor ,method ,value ,@(if daemon (list daemon))))

(defun setmethod-function (flavor method value &optional daemon)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (if daemon
      (if (symbolp value)
	  (let ((mc (getmethod flavor method)))
	    (unless mc
	      (setf (flavor-operations flavor) nil)
	      (setq mc (make-method :name method :flavor flavor))
	      (setmethod-function flavor method mc))
	    (setdaemon (mc daemon) value))
	  (error "Method Object: ~S for (:METHOD ~S ~S ~S) should be a symbol."
		 value
		 (flavor-name flavor) 
		 daemon
		 method))
      (if (or (typep value 'method) (null value))
	  (setf (gethash method (flavor-methods flavor)) value)
	  (error "Method Object: ~S for (:METHOD ~S ~S) should be a method struct."
		 value
		 (flavor-name flavor) 
		 method))))

(defmacro undefmethod ((type daemon &optional method) &rest ignore)
  (declare (ignore ignore))
  (unless method (setq method daemon daemon :primary))
  (let ((flavor (flavor-descriptor type)))
    (let ((m (getmethod flavor method)))
      (when m
	(let ((d (getdaemon m daemon)))
	  (when d
	    (setdaemon (m daemon) nil)
	    (unless (or (getdaemon m :BEFORE)
			(getdaemon m :AFTER)
			(getdaemon m :PRIMARY)
			(getdaemon m :WHOPPER)
			(getdaemon m :WRAPPER))
	      (setmethod (flavor method) nil))))))))

(defmacro method (&rest args) `(method-symbol '(:METHOD ,@args)))

(defun method-symbol (token &optional (failure-result *FAILURE-RESULT*))
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((flavor nil) message daemon)
    (cond 
      ((and (consp token) (eq (first token) :METHOD))
       (setq flavor (eflavor-descriptor (second token))
	     message (if (fourth token) (fourth token) (third token))
	     daemon  (if (fourth token) (third token)  :PRIMARY)))
      ((method-token-p token)
       (setq flavor  (eflavor-descriptor (method-token-name token))
	     message (method-token-message token)
	     daemon (method-token-daemon  token))))
    (cond (flavor (getmethod flavor message daemon))
	  ((eq failure-result :ERROR) (error "Token: ~S should be method specifier" token))
	  ((eq failure-result :TOKEN) token)
	  (T failure-result))))

(defun method-source-code (object type want-list)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((apollo::*GET-SOURCE-FILE-HOOK* nil))
    (apollo:get-source-file (method-symbol object object) type want-list)))

(setq apollo::*GET-SOURCE-FILE-HOOK* #'method-source-code)

(defun get-method-name (flavor daemon message)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (or (getmethod flavor message daemon)
      (gentemp (format nil "MESSAGE-~A-" (get-universal-time)))))
	  
(defmacro make-method-list () `(make-hash-table :test #'eq))
   
(defun mapmethods (function flavor)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (unless (flavor-p flavor) (setq flavor (flavor-descriptor flavor)))
  (when (flavor-methods flavor)
    (maphash function (flavor-methods flavor))))

(defun calculate-combination-list (flavor)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (labels 
      ((components-recurse (flavor) 
	 (cons flavor (mapcan #'components-recurse (copy-list (flavor-components flavor))))))
    (let ((used (components-recurse flavor)))
      (unless (flavor-no-vanilla-flavor flavor)
	(if used
	    (progn (rplacd (last used) (list (flavor-descriptor 'vanilla-flavor))) used)
	    (list (flavor-descriptor 'vanilla-flavor))))
      (do ((u used (cdr u))) 
	  ((endp u) (remove-duplicates used :from-end t))
	(dolist (i (flavor-included-flavors (car u)))
	  (unless (member i (cdr u) :key #'flavor-included-flavors :test #'member)
	    (let ((rest (cdr u)))
	      (setf (cdr u) (calculate-combination-list (eflavor-descriptor i)))
	      (setf (cdr (last u)) rest))))))))

(defun adjust-combination-list (flavor)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((new-combination-list (delete flavor (calculate-combination-list flavor)))
	(*TIME* :ADJUST-COMBINATION-LIST))
    (unless (equal (flavor-combination-list flavor) new-combination-list)
      (setf (flavor-combination-list flavor) new-combination-list)
      (dolist (f (flavor-dependants flavor)) (adjust-combination-list f))))
  (flavor-combination-list flavor))

(defun parse-defflavor-options (flavor-name options variables)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (flet ((check-list-option (option predicate next)
	   (do ((p (cdr option)))
	       ((endp p) (cdr option))
	     (let ((n (funcall next p)))
	       (if (funcall predicate (car p))
		   (setq p n)
		   (progn
		     (warn ";;; For Flavor ~S, Ignored ~S: ~S." flavor-name (car option) (car p))
		     (if (consp n)
			 (setf (car p) (car n) (cdr p) (cdr n))
			 (setq p nil)))))))
	 (echeck-list-option (option predicate next)
	   (do ((p (cdr option) (funcall next p)))
	       ((endp p) (cdr option))
	     (unless (funcall predicate (car p))
	       (error ";;; For Flavor ~S, Bad ~S: ~S" flavor-name (car option) (car p)))))
	 (nonlocal-instance-variable-p (iv)
	   (and iv (symbolp iv) (not (find iv variables :key #'iv-name))))
	 (local-instance-variable-p (iv)
	   (and iv (symbolp iv) (find iv variables :key #'iv-name)))
	 (instance-variable-keyword-p (k)
	   (and (keywordp k)
		(find k variables 
		      :key #'(lambda (v) (symbol-name (iv-name v)))
		      :test #'string-equal)))
	 (flavor-name-p (n) (symbolp n))
	 (method-name-p (n) (keywordp n)))
    (let (combination-modifier combination-type default-init-plist 
			       gettable-instance-variables included-flavors initable-instance-variables
			       init-keywords required-flavors 
			       required-init-keywords required-instance-variables 
			       required-methods settable-instance-variables no-vanilla-flavor)
      (setq combination-type :daemon combination-modifier :base-flavor-last)
      (dolist (o options)
	(if (consp o)
	    (ecase (car o)
	      (:required-instance-variables
	       (setq required-instance-variables (check-list-option o #'nonlocal-instance-variable-p #'cdr)))
	      (:default-init-plist 
		  (setq default-init-plist (check-list-option o #'instance-variable-keyword-p #'cddr)))
	      (:required-init-keywords
	       (setq required-init-keywords (check-list-option o #'instance-variable-keyword-p #'cdr)))
	      (:included-flavors
	       (setq included-flavors (echeck-list-option o #'flavor-name-p #'cdr)))
	      (:required-flavors
	       (setq required-flavors (echeck-list-option o #'flavor-name-p #'cdr)))
	      (:required-methods
	       (setq required-methods (check-list-option o #'method-name-p #'cdr)))
	      (:settable-instance-variables
	       (let ((ivs (check-list-option o #'local-instance-variable-p #'cdr)))
		 (dolist (v ivs) 
		   (pushnew v settable-instance-variables)
		   (pushnew v initable-instance-variables) 
		   (pushnew v gettable-instance-variables))))
	      ((:initable-instance-variables :init-keywords)
	       (let ((ivs (check-list-option o #'local-instance-variable-p #'cdr)))
		 (dolist (v ivs)
		   (pushnew v initable-instance-variables))))
	      (:gettable-instance-variables
	       (let ((ivs (check-list-option o #'local-instance-variable-p #'cdr)))
		 (dolist (v ivs)
		   (pushnew v gettable-instance-variables))))
	      (:method-combination 
	       (let ((type (cadr o))
		     (modifier (caddr o)))
		 (if type     (setf combination-type type))
		 (if modifier (setf combination-modifier modifier)))))
	    (ecase o
	      (:default-init-plist (setq default-init-plist nil))
	      (:no-vanilla-flavor (setf no-vanilla-flavor t))
	      (:settable-instance-variables
	       (dolist (v variables)
		 (let ((o (iv-name v))) 
		   (pushnew o initable-instance-variables) 
		   (pushnew o settable-instance-variables)
		   (pushnew o gettable-instance-variables))))
	      (:initable-instance-variables 
	       (dolist (v variables) 
		 (pushnew (iv-name v) initable-instance-variables)))
	      (:gettable-instance-variables 
	       (dolist (v variables) 
		 (pushnew (iv-name v) gettable-instance-variables))))))
      (setq initable-instance-variables (nreverse initable-instance-variables)
	    gettable-instance-variables (nreverse gettable-instance-variables)
	    settable-instance-variables (nreverse settable-instance-variables)
	    init-keywords (mapcar #'keyword initable-instance-variables))
      (values combination-modifier combination-type default-init-plist 
	      gettable-instance-variables included-flavors initable-instance-variables
	      init-keywords required-flavors 
	      required-init-keywords required-instance-variables 
	      required-methods settable-instance-variables no-vanilla-flavor))))

(defmacro defflavor (name variables components &rest options)
  (defflavor-function name variables components options))

(defun check-flavor-name (name)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (unless (symbolp name) (error "Flavor name: ~S must be a symbol" name)))

(defun check-flavor-variables (name variables)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (unless (or (null variables)
	      (and (consp variables)
		   (do ((v variables (cdr v)))
		       ((null v) t)
		     (unless (and (consp v)
				  (or (symbolp (car v))
				      (and (consp (car v))
					   (cdar v) 
					   (null (cddar v)))))
		       (setq variables v) 
		       (return  nil)))))
    (error "Illegal variable list for flavor: ~S at ~S" name variables)))

(defun check-flavor-components (name components)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (dolist (c components)
    (unless (flavor-descriptor c)
      (error "Unknown Flavor in component List of ~S: ~S" name c))))

(defun check-flavor-options (name options)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (dolist (o options)
    (unless (member (iv-name o)
		    '(:method-combination
		      :default-init-plist 
		      :gettable-instance-variables 
		      :included-flavors 
		      :initable-instance-variables
		      :init-keywords
		      :required-flavors 
		      :required-init-keywords
		      :required-instance-variables 
		      :required-methods
		      :settable-instance-variables :no-vanilla-flavor))
      (error "Illegal Option: ~S for flavor: ~S" o name))))
      
(defun defflavor-type (name)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((subtype (gensym)) (instance (gensym)))
    `((defun ,subtype (,instance) 
	(and
	  (instancep ,instance)
	  (or (eq (flavor-name (instance-flavor ,instance)) ',name)
	      (find (flavor-descriptor ',name) (flavor-combination-list (instance-flavor ,instance))))
	  t))
      (deftype ,name () '(satisfies ,subtype)))))

(defun defflavor-methods (name gettable settable)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((methods nil) (new-value (gensym)))
    (dolist (s gettable)
      (let ((method (keyword (iv-name s))))
	(push `(defmethod-new (,name ,method) () ,s) methods)))
    (dolist (s settable)
      (let ((method (keyword (intern (format nil "SET-~S" (iv-name s))))))
	(push 
	  `(defmethod-new (,name ,method) (,new-value) (setf ,s ,new-value))
	  methods)))
    methods))
  
(defun defflavor-function (name variables components options)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (check-flavor-name name)
  (check-flavor-variables name variables)
  (check-flavor-components name components)
  (check-flavor-options name options)
  (multiple-value-bind
	(combination-modifier combination-type default-init-plist 
			       gettable-instance-variables included-flavors initable-instance-variables
			       init-keywords required-flavors 
			       required-init-keywords required-instance-variables 
			       required-methods settable-instance-variables no-vanilla-flavor)
      (parse-defflavor-options name options variables)
    `(eval-when (load eval compile)
       (defflavor-runtime 
	   ',name ',variables ',components
	   ',combination-modifier ',combination-type ',default-init-plist 
	   ',gettable-instance-variables
	   ',included-flavors 
	   ',initable-instance-variables
	   ',init-keywords ',required-flavors 
	   ',required-init-keywords ',required-instance-variables 
	   ',required-methods 
	   ',settable-instance-variables
	   ',no-vanilla-flavor)
       (eval-when (load eval)
	 ,@(defflavor-type name)
	 ,@(defflavor-methods name gettable-instance-variables settable-instance-variables))
       (pushnew ',name *ALL-FLAVOR-NAMES*)
       ',name)))

(defun defflavor-runtime (name local-variables components
			       combination-modifier combination-type default-init-plist 
			       gettable-instance-variables included-flavors initable-instance-variables
			       init-keywords required-flavors 
			       required-init-keywords required-instance-variables 
			       required-methods settable-instance-variables no-vanilla-flavor)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let* ((f (or (flavor-descriptor name) (make-flavor :name name)))
	 (ocomb-list (flavor-combination-list f)))
    (dolist (u ocomb-list)
      (setf (flavor-dependants u) (delete f (flavor-dependants u))))
    (setf
      (flavor-descriptor name)                   f

      (flavor-name f)                            name
      (flavor-local-variables f)	         local-variables
      (flavor-components f)                      (mapcar #'eflavor-descriptor components)
      (flavor-method-combination-modifier f)     combination-modifier
      (flavor-method-combination-type f)         combination-type
      (flavor-default-init-plist f)              default-init-plist
      (flavor-gettable-instance-variables f)     gettable-instance-variables
      (flavor-initable-instance-variables f)     initable-instance-variables
      (flavor-included-flavors f)                included-flavors
      (flavor-init-keywords f)                   init-keywords
      (flavor-required-flavors f)                required-flavors
      (flavor-required-init-keywords f)          required-init-keywords
      (flavor-required-instance-variables f)     required-instance-variables
      (flavor-required-methods f)                required-methods
      (flavor-settable-instance-variables f)     settable-instance-variables
      (flavor-no-vanilla-flavor f)	         no-vanilla-flavor

      (flavor-all-initable-instance-variables f) nil
      (flavor-all-init-keywords f)               nil
      (flavor-all-required-init-keywords f)      nil
      (flavor-all-required-instance-variables f) nil
      (flavor-all-required-flavors f)            nil
      (flavor-all-required-methods f)            nil
      (flavor-operations f)                      nil

      (flavor-methods f)                         (or (flavor-methods f) (make-method-list))
      (flavor-combination-list f)                (adjust-combination-list f))
    (adjust-variable-list f)
    (dolist (c (flavor-combination-list f)) (pushnew f (flavor-dependants c)))
    (unless (equal (flavor-combination-list f) ocomb-list)
      (let ((*TIME* :DEFFLAVOR)
	    (messages (which-operations f))
	    (affected-flavors (remove-duplicates `(,f ,@(flavor-dependants f) ,@ocomb-list))))
	(fix-combined-method affected-flavors messages :PROPOGATE)
	(fix-combined-method f messages)))))

(defun adjust-flavor-indices (flavor variables)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  #+HASH-FLAVORS #+VECTOR-FLAVORS
  (let ((n 0))
    (setf (flavor-indices flavor) (make-hash-table :test #'eq :size (length variables)))
    (dolist (v variables)
      (setf (gethash v (flavor-indices flavor)) n)
      (incf n)))
  #+PROGV-FLAVORS
  (setf (flavor-indices flavor) (mapcar #'variable-to-reference variables)))

(defun adjust-variable-list (flavor)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((all-variables nil))
    (dolist (s (append 
		 (flavor-local-variables flavor)
		 (mapcan #'(lambda (f) (copy-list (flavor-all-variables f))) (flavor-components flavor))))
      (let ((m (member (iv-name s) all-variables :key #'iv-name)))
	(if m
	    (when (and (consp s) (not (consp (car m)))) (rplaca m s))
	    (push s all-variables))))
    (setq all-variables (nreverse all-variables))
    (unless (equal all-variables (flavor-all-variables flavor))
      (setf (flavor-apply-inside-self flavor) nil)
      (setf (flavor-all-variables flavor) all-variables)
      (let ((variables (mapcar #'iv-name all-variables)))
	(setf (flavor-instance-variables flavor) variables)
	(adjust-flavor-indices flavor variables)
	(dolist (f (flavor-dependants flavor)) (adjust-variable-list f))))
    all-variables))

(defun flavors-with-subcomponent (flavors)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (unless (or (null flavors) (consp flavors)) (setq flavors  (list flavors)))
  (let ((used nil) (used-flavors nil) (some-flavors flavors))
    (loop
      (dolist (f some-flavors)
	(dolist (u (flavor-dependants f))
	  (unless (or (member u used) (member u flavors) (member u used-flavors))
	    (push u used))))
      (if used 
	  (setq used-flavors (append used-flavors used) some-flavors used used nil)
	  (return used-flavors)))))
  
(defun flavor-allowed-init-keywords (name)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (copy-list 
    (combination-slot 
      (eflavor-descriptor name) flavor-all-init-keywords flavor-init-keywords)))

(defun flavor-allowed-init-keywords-p (name key)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (find key (combination-slot 
      (eflavor-descriptor name) flavor-all-init-keywords flavor-init-keywords)))

(defmacro continue-whopper-all ()                `(if *WHOPPER-CONTINUE-HOOK*
						      (funcall *WHOPPER-CONTINUE-HOOK*)))
(defmacro continue-whopper (&rest args)          `(progn 
						    (setq *WHOPPER-ARGS* ,(cons 'list args))
						    (continue-whopper-all)))
(defmacro lexpr-continue-whopper (&rest args)    `(progn 
						    (setq *WHOPPER-ARGS* ',(lexpr-args args))
						    (continue-whopper-all)))

(defmacro lexpr-send (instance message &rest args)
  `(send ,instance ,message ,@(lexpr-args args)))

(defun send (SELF message &rest args)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (special SELF))
  (cond 
    ((instancep SELF) (apply #'flavor-send SELF message args))
    ((fboundp SELF)   (apply SELF message args))
    (t (error "SELF should be an instance: ~S" SELF))))

(defun flavor-send (SELF message &rest args)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (special SELF))
  (with-send-environment (SELF)
    (let ((*WHOPPER-CONTINUE-HOOK* nil)
	  (*WHOPPER-ARGS* args)
	  (combined (or (getmethod (instance-flavor SELF) message :COMBINED)
			(let ((*TIME* :SEND))
			  (fix-combined-method (instance-flavor SELF) message)
			  (getmethod (instance-flavor SELF) message :COMBINED)))))
      (cond (combined (apply combined args))
	    ((operation-handled-p SELF :UNCLAIMED-MESSAGE)
	     (apply #'flavor-send SELF :UNCLAIMED-MESSAGE message args))
	    (t (unclaimed-message SELF message args))))))

(setq lucid::*FLAVORS-SEND* #'send)

(defun check-for-make-instance (flavor options)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let 
      ((component-names (mapcar #'flavor-name (cons flavor (flavor-combination-list flavor))))
       (keywords (do ((l nil) (p options (cddr p))) ((endp p) l) (push (car p) l)))
       (errors nil))
    (let 
	((fk 
	  (set-difference
	    (combination-slot flavor flavor-all-required-init-keywords flavor-required-init-keywords)
	    keywords
	    :test #'eq))
	 (ff 
	  (set-difference
	    (combination-slot flavor flavor-all-required-flavors flavor-required-flavors)
	    component-names
	    :test #'eq))
	 (fm 
	  (set-difference
	    (combination-slot flavor flavor-all-required-methods flavor-required-methods)
	    (which-operations flavor)
	    :test #'eq))
	 (fv 
	  (set-difference 
	    (combination-slot flavor flavor-all-required-instance-variables flavor-required-instance-variables)
	    (flavor-instance-variables flavor)
	    :test #'eq))
	 (fe 
	  (set-difference 
	    keywords 
	    (combination-slot flavor flavor-all-init-keywords flavor-init-keywords)
	    :test #'eq)))
      (when fk (push (format nil "Missing Required Init Keywords: ~S~%"     fk) errors))
      (when ff (push (format nil "Missing Required Flavors: ~S~%"           ff) errors))
      (when fm (push (format nil "Missing Requred Methods: ~S~%"            fm) errors))
      (when fv (push (format nil "Missing Requred Instance Variables: ~S~%" fv) errors))
      (when fe (push (format nil "Unknown Keywords in Options: ~S"          fe) errors))
      (when errors (error "Could Not Make Instance of Flavor: ~S~%~{~A~}" (flavor-name flavor) errors)))))

(defun make-instance (name &rest options)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let* ((flavor (eflavor-descriptor name))
	 (SELF (%create-instance)))
    (declare (special SELF))
    (check-for-make-instance flavor options)
    (setf (instance-flavor SELF) flavor)
    (setf (instance-number SELF) (flavor-instance-number flavor))
    (incf (flavor-instance-number flavor))
    (with-send-environment (SELF)
      (initialize-variables SELF flavor options)
      (when *SEND-INIT-MESSAGE-P* (send SELF :send-if-handles :init)))
    SELF))

(defun combined-method-daemons (flavor message)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((before nil) (whopper nil) (after nil) (primary nil) (wrapper nil)
	(combined (getmethod flavor message :combined)))
    (when combined (setq combined `(,combined)))
    (dolist (f (cons flavor (flavor-combination-list flavor)))
      (let ((m (getmethod f message)))
	(when m
	  (if (method-whopper m) (push (method-whopper m) whopper))
	  (if (method-wrapper m) (push (method-wrapper m) wrapper))
	  (if (method-after   m) (push (method-after   m) after))
	  (if (method-before  m) (push (method-before  m) before))
	  (if (method-primary m) (push (method-primary m) primary)))))
    `(:combined ,combined
		:whopper ,whopper :wrapper ,wrapper :before ,before :primary ,primary :after ,after)))

(defun daemon-combined-method (flavor message daemons)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((before  (nreverse (getf daemons :before)))
	(after   (getf daemons :after))
	(whopper (getf daemons :whopper))
	(wrapper (getf daemons :wrapper))
	(primary (last (getf daemons :primary))))
    (let ((l (+ (length before) (length after) (length primary))))
      (cond ((and (null whopper) (eq l 1))
	     (values 
	       (car (or after before primary))
	       nil))
	    ((and (null whopper) (eq l 0))
	     (values nil nil))
	    (t
	     (let ((func (get-method-name flavor :combined message)))
	       (values 
		 `(defun ,func (&rest *WHOPPER-ARGS*)
		    (declare (special *WHOPPER-ARGS* SELF))
		    ,(let ((a (mapcar #'(lambda (a) `(apply #',a *WHOPPER-ARGS*)) after))
			   (b (mapcar #'(lambda (b) `(apply #',b *WHOPPER-ARGS*)) before))
			   (w (mapcar #'(lambda (w) `(apply #',w *WHOPPER-ARGS*)) whopper))
			   (p (if primary `(apply #',(car primary) *WHOPPER-ARGS*)))
			   (r wrapper))
		       (let ((form (if p (if a `(multiple-value-prog1 ,p ,@a) p) (if a `(progn ,@a nil)))))
			 (if b (setq form (cons 'progn (append b (list form)))))
			 (dolist (wrapper r) (setq form `(,wrapper *WHOPPER-ARGS* ,@form)))
			 (dolist (whopper w)
			   (setq form `(let ((*WHOPPER-CONTINUE-HOOK* (function (lambda () ,form)))) 
					 (declare (special *WHOPPER-CONTINUE-HOOK*))
					 ,whopper)))
			 form)))
		 func)))))))

(defun hairy-combined-method (flavor message daemons operation)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let* ((primary (getf daemons :primary))
	 (l (length primary)))
    (cond ((and (eq l 1) (member operation '(progn and or append nconc)))
	   (values (car primary) nil))
	  ((eq l 0)
	   (values nil nil))
	  (t
	   (let ((func (get-method-name flavor :combined message)))
	     (values 
	       `(defun ,func (&rest *WHOPPER-ARGS*)
		  (declare (special *WHOPPER-ARGS* SELF))
		  (,operation
		     ,@(mapcar #'(lambda (p) `(apply #',p *WHOPPER-ARGS*)) primary)))
	       func))))))

(defun build-combined-method (flavor message)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((daemons (combined-method-daemons flavor message)))
    (case (flavor-method-combination-type flavor)
      (:DAEMON (daemon-combined-method flavor message daemons))
      (:PROGN  (hairy-combined-method flavor message daemons 'progn))
      (:AND    (hairy-combined-method flavor message daemons 'and))
      (:OR     (hairy-combined-method flavor message daemons 'or))
      (:LIST   (hairy-combined-method flavor message daemons 'list))
      (:APPEND (hairy-combined-method flavor message daemons 'append))
      (:NCONC  (hairy-combined-method flavor message daemons 'nconc))
      (:CASE (error "Not Implemented"))
      ;; The following will probably not be implemented
      (:INVERSE-LIST (error "Not Implemented"))
      (:PASS-ON (error "Not Implemented"))
      (:DAEMON-WITH-OR (error "Not Implemented"))
      (:DAEMON-WITH-AND (error "Not Implemented"))
      (:DAEMON-WITH-OVERIDE (error "Not Implemented")))))

(defmacro defmethod-let ((type SELF) vars . body) 
  (declare (ignore type SELF vars))
  body)
     
(defun defmethod-function (name daemon message args body)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((flavor (eflavor-descriptor name))
	(bv nil))
    (labels ((iv-or-variable (iv)
	       (if (or (member iv (flavor-instance-variables flavor) :test #'eq)
		       (member iv (flavor-required-instance-variables flavor) :test #'eq))
		   `(getvariable SELF ,iv)
		   iv))
	     (iv-or-variable-dynamic (iv)
	       (if (or (member iv (flavor-instance-variables flavor) :test #'eq)
		       (member iv (flavor-required-instance-variables flavor) :test #'eq))
		   `(getvariable-from-instance SELF ,iv)
		   iv))
	     (send-special-action (form)
	       (apollo:expand `(flavors::flavor-send ,@(cdr form))))
	     (setq-special-action (form)
	       (declare (special apollo:*BOUND-VARIABLES* apollo:*GLOBALS-SET*))
	       (do ((f (cdr form) (cddr f))
		    (result (list 'setf)))
		   ((null f) (nreverse result))
		 (let ((set-variable (car f))
		       (set-to (cadr f)))
		   (if (member set-variable apollo:*BOUND-VARIABLES* :test #'eq)
		       (push set-variable result)
		       (progn 
			 (push (iv-or-variable set-variable) result)
			 (pushnew set-variable apollo:*GLOBALS-SET*)))
		   (push (apollo:expand set-to) result))))
	     (multiple-value-bind-special-action (form)
	       `(LISP::MULTIPLE-VALUE-BIND ,@(cdr (apollo:expand `(lambda ,@(cdr form)))))))
      (macrolet ((ivs (instance) `(instance-variables ,instance)))
	(let ((apollo:*BOUND-VARIABLES* (list SELF))
	      (apollo:*special-actions* 
	       (list (list 'LISP:MULTIPLE-VALUE-BIND #'multiple-value-bind-special-action)
		     (list 'LISP:SETQ                #'setq-special-action)))
	      (apollo:*UNBOUND-VARIABLE-ACTION* #'iv-or-variable-dynamic))
	  (setq args (apollo:argument-variables-from-arglist args)
		bv   apollo:*BOUND-VARIABLES*)))
      (multiple-value-bind (code-walk-body code-walk-info)
	  (apollo:code-walk 
	    `(lambda nil ,@body)
	    :BOUND-VARIABLES bv
	    :SPECIAL-ACTIONS (list (list 'LISP:MULTIPLE-VALUE-BIND #'multiple-value-bind-special-action)
				   (list 'LISP:SETQ                #'setq-special-action)
				   (list 'flavors::send            #'send-special-action)
				   )
	    :UNBOUND-VARIABLE-ACTION #'iv-or-variable)
	(let* ((func-name (get-method-name flavor daemon message))
	       (body #-(or HASH-FLAVORS VECTOR-FLAVORS) body
		     #+(or HASH-FLAVORS VECTOR-FLAVORS) code-walk-body)
	       (declarations-and-strings 
		(do ((b (cddr body) (cdr b))
		     (d&s nil))
		    ((null b) (progn (setq body nil) (nreverse d&s)))
		  (if (or (stringp (car b)) (and (consp (car b)) (eq (caar b) 'lisp::declare)))
		      (push (car b) d&s)
		      (return (progn (setq body b) (nreverse d&s))))))
	       (globals-used 
		(intersection (union (getf code-walk-info :GLOBALS-ACCESSED)
				     (getf code-walk-info :GLOBALS-SET))
			      (union (flavor-instance-variables flavor)
				     (flavor-required-instance-variables flavor)))))
	  (values `(defun ,func-name ,args
		     ,@declarations-and-strings
		     ,@(with-flavor-ivs globals-used body))
		  func-name))))))

(defmacro defmethod-new ((name daemon &optional method) args &Rest body)
  (unless method (setq method daemon daemon :primary))
  (multiple-value-bind (func func-name) 
      (defmethod-function name daemon method args body)
    `(eval-when (load eval compile)
       (unless (fboundp ',func-name)
	 ,func
	 (setmethod ((flavor-descriptor ',name) ,method ,daemon) ',func-name)
	 (setf (gethash ',func-name *PRINT-NAMES*) 
	       (make-method-token :name ',name :daemon ,daemon :message ,method))))))

(defmacro defmethod ((name daemon &optional method) args &Rest body)
  (unless method (setq method daemon daemon :primary))
  (multiple-value-bind (func func-name) 
      (defmethod-function name daemon method args body)
    `(eval-when (load eval compile)
       ,func
       (setmethod ((flavor-descriptor ',name) ,method ,daemon) ',func-name)
       (setf (gethash ',func-name *PRINT-NAMES*) 
	     (make-method-token :name ',name :daemon ,daemon :message ,method)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
(defun cleanup-all-flavors () nil)

(defun symeval-in-instance (instance variable &optional no-error-p (unbound *UNBOUND*))
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (if (find variable (flavor-instance-variables (instance-flavor instance)) :test #'eq)
      (let ((r (with-flavor-ivs-dynamic (instance variable) (getvariable-dynamic instance variable))))
	(if (eq r *UNBOUND*) unbound r))
      (unless no-error-p
	(error "No instance variable: ~S for ~S" variable instance))))

(defun set-in-instance (instance variable value)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (if (find variable (flavor-instance-variables (instance-flavor instance)) :test #'eq)
      (with-flavor-ivs-dynamic (instance variable)
	(setvariable-dynamic (instance variable) value))
      (error "No instance variable: ~S for ~S" variable instance)))

(defsetf symeval-in-instance (instance symbol &optional no-error-p unbound) (value)
  (declare (ignore no-error-p unbound))
  `(set-in-instance ,instance ,symbol ,value))

(defmacro without-cleaning-flavors (&rest form)
  `(progn ,@form))

(defmacro defwhopper ((flavor method) args &rest body)
  `(defmethod (,flavor :whopper ,method) ,args ,@body))

(defmacro defwrapper ((flavor method) args &rest body)
  `(defmethod (,flavor :wrapper ,method) ,args ,@body))

(defun which-operations (flavor)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (unless (flavor-p flavor) (setq flavor (flavor-descriptor flavor)))
  (or (flavor-operations flavor)
      (let ((messages nil))
	(dolist (f (cons flavor (flavor-combination-list flavor)) messages)
	  (mapmethods #'(lambda (key method) 
			  (declare (ignore method)) 
			  (pushnew key messages)) 
		      f))
	(setf (flavor-operations flavor) messages)
	messages)))

(defun operation-handled-p (SELF message)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (if (instance-flavor SELF)
      (if (find message (which-operations (type-of SELF))) t nil)))

(defun get-handler-for (instance message)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((*TIME* :GET-HANDLER))
    (or (getmethod (instance-flavor instance) message :COMBINED)
	(fix-combined-method (instance-flavor instance) message))))

(defun unclaimed-message (SELF message &rest args)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (declare (special SELF) (ignore args))
  (error "No Method: ~S for ~S" message SELF))

(defun fix-combined-method (flavors messages &optional do-dependants)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (unless (or (null flavors)  (consp flavors))  (setq flavors  (list flavors)))
  (unless (or (null messages) (consp messages)) (setq messages (list messages)))
  (when do-dependants (setq flavors (append flavors (flavors-with-subcomponent flavors))))
  (when (find *TIME* *RESET-COMBINED-WHEN*) 
    (dolist (flavor flavors)
      (dolist (message messages) (setmethod (flavor message :combined) nil))))
  (when (find *TIME* *COMPILE-COMBINED-WHEN*)
    (dolist (flavor flavors)
      (when (or (flavor-instanciated flavor)
		(find *TIME* *COMPILE-UNINSTANCIATED-WHEN*))
	(dolist (message messages)
	  (when (not (getmethod flavor message :combined))
	    (set-combined-method flavor message)))))))

(defun set-combined-method (flavor message)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (multiple-value-bind (m m-name) (build-combined-method flavor message)
    (declare (optimize (compilation-speed 0) (safety 0) (speed 3) (space 0) (size 0)))
    (when m-name
      (eval m)
      (setf (gethash m-name *PRINT-NAMES*) 
	    (make-method-token :name (flavor-name flavor) :daemon :COMBINED :message message))
      (setq m m-name))
    (when m
      (unless (compiled-function-p (symbol-function m)) (compile m))
      (setmethod (flavor message :combined) m))))

(defmacro compile-flavor-methods (&rest names)
  (let ((forms nil))
    (dolist (name names)
      (push 
	`(dolist (message (which-operations (flavor-descriptor name)))
	   (set-combined-method (eflavor-descriptor ',name) message))
	forms))
    `(eval-when (eval load) ,@forms)))

(defun recompile-flavor (name &optional messages (do-dependants t) (use-old-combined-methods t))
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((flavor (eflavor-descriptor name))
	(*TIME* (if use-old-combined-methods :COMPILE :RECOMPILE)))
    (unless messages (setq messages (which-operations flavor)))
    (fix-combined-method flavor messages do-dependants)
    (let ((flavors (cons flavor (flavor-combination-list flavor))))
      (when do-dependants (setq flavors (append flavors (flavors-with-subcomponent flavor))))
      (dolist (f flavors)
	(dolist (message messages) 
	  (let ((method (getmethod f message)))
	    (when method 
	      (dolist (d (list (method-wrapper  method)
			       (method-whopper  method)
			       (method-before   method)
			       (method-primary  method)
			       (method-after    method)))
		(when (and d (not (compiled-function-p (symbol-function d))))
		  (compile d))))))))))

(defun flavor-from-spec (spec)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (if (flavor-p spec)
      spec
      (if (instancep spec)
	  (instance-flavor spec)
	  (if (symbolp spec)
	      (eflavor-descriptor spec)
	      (error "Illegal flavor specifier: ~S" spec)))))
    
(defun flavor-tree (flavor branch-function &optional insert-function ancestors)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let ((branch (funcall branch-function flavor)))
    (if (and branch (not (member flavor ancestors)))
	(let ((ancestors (cons flavor ancestors)))
	  (cons (funcall insert-function flavor)
		(mapcar #'(lambda (f) (flavor-tree f branch-function insert-function ancestors)) branch)))
	(if insert-function (funcall insert-function flavor) flavor))))

(defun show-dependants (flavor-name-or-instance)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (flavor-tree (flavor-from-spec flavor-name-or-instance)
	       #'(lambda (f) (flavor-dependants f))
	       #'(lambda (f) (flavor-name f))))

(defun show-components (flavor-name-or-instance)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (flavor-tree (flavor-from-spec flavor-name-or-instance)
	       #'(lambda (f) (flavor-components f))
	       #'(lambda (f) (flavor-name f))))

(defun show-instance-variables (flavor-name-or-instance &key match (local nil))
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (let* ((flavor (flavor-from-spec flavor-name-or-instance))
	 (variables (mapcar #'iv-name 
			    (if local 
				(flavor-local-variables flavor)
				(flavor-all-variables flavor)))))
    (if match
	(progn
	  (when (symbolp match) (setq match (symbol-name match)))
	  (mapcan #'(lambda (v) (when (search match (symbol-name v)) (list v))) variables))
	variables)))

(defun show-methods (flavor &key 
			    match 
			    (local nil) 
			    (messages nil)
			    (daemons '(:BEFORE :AFTER :PRIMARY :WHOPPER :WRAPPER)))
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (setq flavor (flavor-from-spec flavor))
  (unless messages (setq messages (which-operations flavor)))
  (let* ((name (flavor-name flavor))
	 (methods
	  (mapcan #'(lambda (message) 
		      (mapcan #'(lambda (d) 
				  (when (consp d) 
				    (case (caddr (car d))
				      (:primary (setq d (last d)))
				      (:before  (setq d (nreverse d))))
				    (mapcar #'(lambda (m) (rplaca m :METHOD)) d)))
			      (combined-method-daemons flavor message)))
		  messages)))
    (setq methods (delete-if-not #'(lambda (e) (member e daemons)) methods :key #'caddr))
    (if local (setq methods (delete-if-not #'(lambda (e) (eq e name)) methods :key #'cadr)))
    (when match
      (when (symbolp match) (setq match (symbol-name match)))
      (setq methods (delete-if-not #'(lambda (e) (search match (symbol-name e))) methods :key #'cadddr)))
    methods))

(defun define-apply-inside-self (SELF)
  (declare (optimize (speed 3) (space 0) (safety 1) (compilation-speed 0)))
  (setf (flavor-apply-inside-self (instance-flavor SELF))
	(symbol-function
	  (let* ((func (gensym)) (args (gensym)) (name (gensym)) (ivariables (gensym)) 
		 (flavor (instance-flavor SELF))
		 (variables (flavor-instance-variables flavor))
		 (reset        (mapcar #'(lambda (s) `(setvariable (,ivariables ',s) ,s)) variables))
		 (getvariables (mapcan #'(lambda (s) `(,s (getvariable  ,ivariables ,s))) variables))
		 (f `(function (lambda (,func ,args)
		       (let ((,name (type-of SELF))
			     (,ivariables (instance-variables SELF)) ,@variables)
			 (declare (special SELF *INSIDE* ,@variables))
			 (declare (inline gethash set-gethash))
			 (unwind-protect 
			      (progn
				(unless (eq *INSIDE* SELF) (setq ,@getvariables))
				(let ((*INSIDE* SELF)) (apply ,func ,args)))
			   (unless (eq *INSIDE* SELF) ,@reset)))))))
	    (compile func f)
	    func))))


