; -*- Mode: Common-Lisp; Package: Tools -*-
(in-package 'tools)
#+LUCID (file-proclaim '(optimize (speed 3) (safety 1) (compilation-speed 0) (space 0)))
(export '(nset-difference-order-preserving set-difference-order-preserving do-recurse
	  unionf set-differencef *UNBOUND* maxf minf
	  deletef letf letf*))

(defvar *UNBOUND* (intern "<UNBOUND>"))

(defun nset-difference-order-preserving (s s2 &key (test #'eq) (key #'identity))
  (let ((r nil)
	(rlast nil))
    (loop
      (if (endp s) 
	  (return r)
	  (if (member (car s) s2 :test test :key key)
	      (pop s)
	      (if rlast
		  (setf (cdr rlast) s s (cdr s) rlast (cdr rlast) (cdr rlast) nil)
		  (setf rlast s r rlast s (cdr s) (cdr rlast) nil)))))
    ))

(defun nintersection-order-preserving (s s2 &key (test #'eq) (key #'identity))
  (let ((r nil)
	(rlast nil))
    (loop
      (if (endp s) 
	  (return r)
	  (if (member (car s) s2 :test test :key key)
	      (if rlast
		  (setf (cdr rlast) s s (cdr s) rlast (cdr rlast) (cdr rlast) nil)
		  (setf rlast s r rlast s (cdr s) (cdr rlast) nil))
	      (pop s))))))

(defun nunion-order-preserving (s s2)
  (delete-duplicates (nconc s s2)))

;(defmacro do-recurse ((next)
;		      (curr list 
;			    &optional result
;			    &key 
;			    (destructive-p nil)
;			    (discipline :DEPTH-FIRST)
;			    (test #'eq)
;			    (key #'identity)
;			    return-done-p
;			    recursive-continue-flag)
;		      &rest body)
;  (let ((todo (gensym)) (done (gensym)))
;    `(let (,curr 
;	   ,@(if recursive-continue-flag `(,recursive-continue-flag)))
;       (do ((,done nil)
;	    (,todo 
;	       (if (listp ,list) ,(if destructive-p list `(copy-list ,list)) (list ,list))
;	       ,(let ((form 
;		       `(nunion-order-preserving 
;			  (nset-difference-order-preserving 
;			    ,(if destructive-p next `(copy-list ,next)) 
;			    ,done
;			    :test ,test :key ,key)
;			  ,todo)))
;		  (if recursive-continue-flag
;		      `(if ,recursive-continue-flag ,form ,todo)
;		      form))))
;	   ((endp ,todo) ,(if return-done-p done result))
;	 (setq ,curr (pop ,todo)
;	       ,@(if recursive-continue-flag `(,recursive-continue-flag t)))
;	 (push ,curr ,done)
;	 ,@body))))

(defmacro do-recurse ((next-function)
		      (curr next
			    &optional result
			    &rest keys)
		      &rest body)
  (let ((form `(do-recurse-function 
		   ,next
		 (function ,(if (consp next-function) `(lambda (,curr) ,next-function) next-function))
		 ,(when body `(function (lambda (,curr) ,@body)))
		 ,@keys)))
    (case result ((nil) form) (:VISITED `(nreverse ,form)) (t `(progn ,form ,result)))))

(defun do-recurse-function (current func body
				 &key 
				 (discipline :BASE-FLAVOR-LAST )
				 (test #'eq)
				 (key #'identity)
				 (execute-when #'identity)
				 (execute-once-only-p nil))
  (let ((visited nil))
    (dolist (current (if (consp current) current (list current)) visited)
      (if execute-once-only-p
	  (progn (when body (funcall body current)) (list current))
	  (labels ((BASE-FLAVOR-LAST (current)
		     (unless (member current visited :test test :key key)
		       (push current visited)
		       (when (funcall execute-when current)
			 (when body (funcall body current))
			 (dolist (b (funcall func current)) (BASE-FLAVOR-LAST b))))))
	    (ecase discipline (:BASE-FLAVOR-LAST (base-flavor-last current))))))))

(defmacro deletef (item place &rest keys)
  `(setf ,place (delete ,item ,place ,@keys)))

(define-modify-macro unionf (&rest args) union)
(define-modify-macro set-differencef (&rest args) set-difference)
(define-modify-macro maxf (&rest args) max)
(define-modify-macro minf (&rest args) max)

(defmacro letf (elements &rest body)
  (let ((slots nil) (old-values nil) (new-values nil))
    (dolist (e elements)
      (if (consp e) 
	  (progn (push (car e) slots) (push (cadr e) new-values) (push (gensym) old-values))
	  (progn (push e       slots) (push nil      new-values) (push (gensym) old-values))))
    (setq slots (nreverse slots) new-values (nreverse new-values))
    `((lambda ,old-values
	(unwind-protect
	     (progn
	       ,@(mapcar #'(lambda (s v) `(setf ,s ,v)) slots new-values)
	       ,@body)
	  (progn ,@(mapcar #'(lambda (s v) `(setf ,s ,v)) slots old-values))))
      ,@slots)))

(defmacro letf* (elements &rest body)
  (cond ((null elements) `(progn ,@Body))
	((= (length elements) 1) `(letf ,elements ,@body))
	(t `(letf (,(car elements)) (letf* ,(cdr elements) ,@Body)))))

(defun nsubseq (array &optional (start 0) end)
  (let ((end (or end (length array))))
    (make-array (- end start)
		:element-type (array-element-type array)
		:displaced-to array
		:displaced-index-offset start)))

(provide 'tools)
    
 