(in-package 'flavors)
(require 'flavors "~/mdw/flavors/flavors")

(defflavor vanilla-flavor () () :no-vanilla-flavor)

(defmethod (vanilla-flavor :describe) (&optional (stream t) &key match (local nil))
  (let ((ivs (show-instance-variables self :match match :local local))
	(l 0))
    (declare (fixnum l)
	     (ftype (function (fixnum fixnum) fixnum) max))
    (dolist (iv ivs) 
      (setq l (max l (the fixnum (length (the string (symbol-name iv)))))))
    (incf l 2)
    (dolist (iv ivs) 
      (format stream "~S:~VT~S~%" iv l (symeval-in-instance self iv)))))

(defmethod (vanilla-flavor :print-self) (&optional (stream *STANDARD-OUTPUT*) prindepth slashify-p)
  (ordinary-print-instance self stream prindepth slashify-p))

(defmethod (vanilla-flavor :which-operations) () (which-operations (type-of self)))

(defmethod (vanilla-flavor :operation-handled-p) (message) (operation-handled-p self message))

(defmethod (vanilla-flavor :send-if-handles) (message &rest args)
  (when (operation-handled-p self message) (apply #'send self message args)))

(defmethod (vanilla-flavor :lexpr-send-if-handles) (message &rest args)
  (when (operation-handled-p self message) (apply #'send self message (lexpr-args args))))
  
(defmethod (vanilla-flavor :get-handler-for) (operation)
  (get-handler-for self operation))

(defmethod (vanilla-flavor :unclaimed-message) (message &rest args)
  (unclaimed-message self message args))

(defmethod (vanilla-flavor :eval-inside-yourself) (form)
  (unless  (flavor-apply-inside-self (instance-flavor self)) (define-apply-inside-self self))
  (funcall (flavor-apply-inside-self (instance-flavor self)) #'eval (list form)))

(defmethod (vanilla-flavor :funcall-inside-yourself) (function &rest args)
  (unless  (flavor-apply-inside-self (instance-flavor self)) (define-apply-inside-self self))
  (funcall (flavor-apply-inside-self (instance-flavor self)) function args))

(defmethod (vanilla-flavor :recompile-flavor) 
    (&optional messages (do-dependants t) (use-old-combined-methods nil))
  (recompile-flavor (flavor-name (instance-flavor self)) messages do-dependants use-old-combined-methods))

(defmethod (vanilla-flavor :show-dependants) ()
  (flavor-tree (flavor-from-spec self)
	       #'(lambda (f) (flavor-dependants f))
	       #'(lambda (f) (flavor-name f))))

(defmethod (vanilla-flavor :show-components) ()
  (flavor-tree (flavor-from-spec self)
	       #'(lambda (f) (flavor-components f))
	       #'(lambda (f) (flavor-name f))))

(defmethod (vanilla-flavor :show-instance-variables) (&key match (local nil))
  (show-instance-variables self :match match :local local))

(defmethod (vanilla-flavor :show-methods) (&key 
					    match 
					    (local nil) 
					    (messages nil)
					    (daemons '(:BEFORE :AFTER :PRIMARY :WHOPPER :WRAPPER)))
  (show-methods self :match match :local local :messages messages :daemons daemons))

(defmethod (vanilla-flavor :component-of) ()
  (mapcar #'(lambda (f) (flavor-name f)) (flavor-dependants (instance-flavor self))))

(defmethod (vanilla-flavor :component-flavors) (&key (local t))
  (mapcar #'(lambda (f) (flavor-name f))
	  (if local
	      (flavor-components       (instance-flavor self))
	      (flavor-combination-list (instance-flavor self)))))

(defmethod (flavors::vanilla-flavor :break) () (break "Break"))